/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2003 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* pid_t */
#include <errno.h>			/* errno */
#include <signal.h>			/* sigaction() */
#include <stdio.h>			/* puts() */
#include <stdlib.h>			/* exit() */
#include <string.h>			/* strcmp() */
#include <unistd.h>			/* fork() */

/* waitpid() */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include "clex.h"
#include "exec.h"

#include "cfg.h"			/* config_str() */
#include "control.h"		/* get_current_mode() */
#include "edit.h"			/* edit_islong() */
#include "inout.h"			/* win_remark() */
#include "history.h"		/* hist_save() */
#include "list.h"			/* list_directory() */
#include "tty.h"			/* tty_setraw() */
#include "util.h"			/* base_name() */
#include "ustring.h"		/* us_copy() */
#include "xterm_title.h"	/* xterm_title_set() */

extern int errno;

#define MAX_SHELL_ARGS	8
static char *shell_argv[MAX_SHELL_ARGS + 1];
static int cmd_index;		/* which parameter is the command
							   to be executed */

void
exec_initialize(void)
{
	US_INIT(line_cmd.line);
	US_INIT(line_cmd.save_line);

	exec_shell_reconfig();
	exec_prompt_reconfig();	/* prompt needs to know the shell */
}

static int
parse_shellprog(const char *shellprog)
{
	char *pch;
	int i;
	FLAG arg_end;
	static USTRING dup = { 0,0 };

	us_copy(&dup,shellprog);

	/* split to arguments */
	for (arg_end = 1, i = 0, pch = USTR(dup); *pch; pch++ )
		if (*pch == ' ') {
			*pch = '\0';
			arg_end = 1;
		}
		else if (TCLR(arg_end)) {
			if (i >= MAX_SHELL_ARGS) {
				txt_printf(
				  "CONFIG: Incorrect SHELLPROG: too many arguments\n"
				  "        limit is %d\n",MAX_SHELL_ARGS);
				return -1;
			}
			shell_argv[i++] = pch;
		}
	shell_argv[i] = 0;

	while (--i >= 1)
		if (strcmp(shell_argv[i],"<COMMAND>") == 0) {
			cmd_index = i;	/* real command belongs here */
			return 0;		/* success */
		}

	txt_printf("CONFIG: Incorrect SHELLPROG: missing \"<COMMAND>\"\n");
	return -1;
}

/* prepare to execute commands according to "SHELLPROG" */
void
exec_shell_reconfig(void)
{
	const char *shellprog;

	shellprog = config_str(CFG_SHELLPROG);
	if (*shellprog) {
		if (parse_shellprog(shellprog) == 0)
			return;		/* success */
		txt_printf("EXEC: Will use your login shell: %s\n",
		  clex_data.shell);
	}

	/* SHELLPROG = AUTO: loginshell -c <COMMAND> */
	((const char **)shell_argv)[0] = clex_data.shell;
	shell_argv[1] = "-c";
	cmd_index = 2;
	shell_argv[3] = 0;
}

void
exec_prompt_reconfig(void)
{
	int len;
	const char *prompt, *chars;
	static USTRING cmd_prompt = { 0,0 };
	static char prompt_char[4] = " ? ";

	prompt = config_str(CFG_PROMPT);

	if (prompt[0] == '\0') {
		/* PROMPT=AUTO */
		prompt = base_name(shell_argv[0]);
		len = strlen(prompt);
		if (len >= 2 && strcmp(prompt + len - 2,"sh") == 0)
			if (len >= 3 && prompt[len - 3] == 'c')
				chars = "%#";	/* looks like C-shell */
			else
				chars = "$#";	/* looks like Bourne shell */
		else
			chars = ">>";		/* some other shell */
		prompt_char[1] = chars[clex_data.isroot];
	}
	else
		prompt_char[1] = '\0';

	us_cat(&cmd_prompt,clex_data.isroot ? "ROOT " : "",
	  prompt,prompt_char,(char *)0);
	edit_setprompt(&line_cmd,USTR(cmd_prompt));
}

static int
execute(const char *command)
{
	pid_t childpid;
	FLAG failed;
	int status, code;
	struct sigaction act;

	failed = 1;
	xterm_title_set(1,command);
	childpid = fork();
	if (childpid == -1)
		printf("EXEC: Cannot create new process (%s)\n",
		  strerror(errno));
	else if (childpid == 0) {
		/* child process = command */
#ifdef _POSIX_JOB_CONTROL
		/* move this process to a new foreground process group */
		childpid = getpid();
		setpgid(childpid,childpid);
		tcsetpgrp(STDIN_FILENO,childpid);
#endif

		/* reset signal dispositions */
		act.sa_handler = SIG_DFL;
		act.sa_flags = 0;
		sigemptyset(&act.sa_mask);
		sigaction(SIGINT,&act,0);
		sigaction(SIGQUIT,&act,0);
#ifdef _POSIX_JOB_CONTROL
		sigaction(SIGTSTP,&act,0);
		sigaction(SIGTTIN,&act,0);
		sigaction(SIGTTOU,&act,0);
#endif

		/* execute the command */
		((const char **)shell_argv)[cmd_index] = command;
		execv(shell_argv[0],shell_argv);
		printf("EXEC: Cannot execute shell %s (%s)\n",
		  shell_argv[0],strerror(errno));
		exit(99);
		/* NOTREACHED */
	}
	else {
		/* parent process = CLEX */
#ifdef _POSIX_JOB_CONTROL
		/* move child process to a new foreground process group */
		setpgid(childpid,childpid);
		tcsetpgrp(STDIN_FILENO,childpid);
#endif
		for (; /* until break */;) {
			/* wait for child process exit or stop */
			while (waitpid(childpid,&status,WUNTRACED) < 0)
				/* ignore EINTR */;
#ifdef _POSIX_JOB_CONTROL
			/* move CLEX to foreground */
			tcsetpgrp(STDIN_FILENO,clex_data.pid);
#endif
			if (!WIFSTOPPED(status))
				break;

			puts(
				"\r\n"
				"\r\n"
				"EXEC: CLEX does not support job control\r\n"
				"EXEC: suspended command will be resumed shortly\r\n"
				"\r\n");
			fflush(stdout);
#ifdef _POSIX_JOB_CONTROL
			/* move command back to foreground */
			tcsetpgrp(STDIN_FILENO,childpid);
#endif
			sleep (2);
			kill(-childpid,SIGCONT);
		}

		tty_reset();
		if (WIFEXITED(status)) {
			code = WEXITSTATUS(status);
			if (code == 0) {
				failed = 0;
				fputs("\nCommand successful. ",stdout);
			}
			else
				printf("\nExit code = %d. ",code);
		}
		else {
			code = WTERMSIG(status);
			printf("\nAbnormal termination, signal %d",code);
#ifdef SYS_SIGLIST_DECLARED
			printf(" (%s)",sys_siglist[code]);
#endif
#ifdef WCOREDUMP
			if (WCOREDUMP(status))
				fputs(", core image dumped",stdout);
#endif
			putchar('\n');
		}
	}

	/*
	 * list directory while the user stares at the screen
	 * reading the command output, CLEX appears to restart
	 * very much faster. All output routines called from
	 * list_directory() must check if curses mode is active,
	 * win_warning() is the only function designed for this
	 * situation.
	 */
	xterm_title_set(0,command);
	list_directory();
	ppanel_file->other->expired = 1;

	tty_press_enter();
	xterm_title_set(0,0);

	return failed;
}

static int
user_confirm(void)
{
	int in, confirmed;

	fputs("\nExecute the command ? (y = YES) ",stdout);
	fflush(stdout);
	tty_setraw();
	in = getchar();
	tty_reset();
	confirmed = in == 'y' || in == 'Y';
	puts(confirmed ? "yes\n" : "no\n");
	fflush(stdout);
	return confirmed;
}

static int
test_word(const char *line, const char *word)
{
	char ch;

	while (*line == ' ')
		line++;
	while ( (ch = *word++) != '\0')
		if (*line++ != ch)
			return 0;
	return 1;
}

/*
 * return value is the warning level:
 *   0 = no warning issued
 *   1 = only mandatory warning(s)
 *   2 = configurable warning(s)
 */
static int
print_warnings(const char *cmd)
{
	static USTRING cwd = { 0,0 };
	int warn;

	warn = 0;		/* 0 = no warning yet */

	if (get_cwd_us(&cwd) < 0) {
		puts("WARNING: current working directory is not accessible");
		warn = 1;	/* 1 = this warning cannot be turned off */
		us_copy(&cwd,"???");
	}
	else if (strcmp(USTR(ppanel_file->dir),USTR(cwd))) {
		printf("WARNING: current working directory has been renamed:\n"
			   "         old name: %s\n"
			   "         new name: %s\n",
		  USTR(ppanel_file->dir),USTR(cwd));
		us_copy(&ppanel_file->dir,USTR(cwd));
		warn = 1;
	}

	if (config_num(CFG_WARN_RM) && test_word(cmd,"rm ")) {
		printf("working directory: %s\n"
			   "WARNING: rm command deletes files, please confirm\n",
			USTR(cwd));
		warn = 2;	/* 2 = this warning can be turned off */
	}

	/* following warnings are not appropriate in history mode */
	if (get_current_mode() == MODE_FILE) {
		if (config_num(CFG_WARN_CD) && test_word(cmd,"cd ")) {
			puts("WARNING: External cd command will NOT "
			  "change the directory in the CLEX");
			warn = 2;
		}
		/*
		 * edit_islong() works with 'textline' instead of 'cmd',
		 * because it is meaningfull only in MODE_FILE
		 */
		if (config_num(CFG_WARN_LONG) && edit_islong()) {
			puts("WARNING: This long command did not fit to the"
			  " command line");
			warn = 2;
		}
	}

	fflush(stdout);
	return warn;
}

/*
 * function returns 1 if the command has been executed
 * (successfully or not), otherwise 0
 */
int
execute_cmd(const char *cmd)
{
	static FLAG hint = 1;
	int do_exec, warn_level;

	curses_stop();
	putchar('\n');
	puts(cmd);
	putchar('\n');
	fflush(stdout);

	warn_level = print_warnings(cmd);
	do_exec = warn_level == 0 || user_confirm();
	if (do_exec)
		hist_save(cmd,execute(cmd));

	curses_restart();
	if (warn_level >= 2 && TCLR(hint))
		win_remark("warnings can be turned off in the config panel");

	return do_exec;
}
