/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2003 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>  /* clex.h */
#include <stdlib.h>		/* qsort() */
#include <string.h>		/* strcmp() */

/* major() */
#ifdef MAJOR_IN_MKDEV
# include <sys/mkdev.h>
#endif
#ifdef MAJOR_IN_SYSMACROS
# include <sys/sysmacros.h>
#endif

#include "clex.h"
#include "sort.h"

#include "cfg.h"		/* config_num() */
#include "directory.h"	/* filepos_save() */
#include "panel.h"		/* pan_adjust() */
#include "sdstring.h"	/* SDSTR() */

static int group_files;
/* result of CMP() is an int even if V1 and V2 are not */
#define CMP(V1,V2) ((V1) == (V2) ? 0 : (V1) < (V2) ? -1 : 1)

void
sort_prepare(void)
{
	panel_sort.top = -1;
	panel_sort.curs = panel_sort.order;
	panel = (PANEL *)&panel_sort;
	pan_adjust(panel);
	textline = 0;
}

void
cx_sort_set(void)
{
	if (panel_sort.curs >= 0 && panel_sort.order != panel_sort.curs) {
		panel_sort.order = panel_sort.curs;
		filepos_save();
		sort_files();
		filepos_set();
		ppanel_file->other->expired = 1;
	}
	next_mode = MODE_SPECIAL_RETURN;
}

/* compare reversed strings */
static int
revstrcmp(const char *s1, const char *s2)
{
	size_t i1, i2;
	int c1, c2;

	for (i1 = strlen(s1), i2 = strlen(s2); i1 > 0 && i2 > 0;) {
		c1 = s1[--i1] & 0xFF;
		c2 = s2[--i2] & 0xFF;
		/*
		 * ignoring LOCALE, this sort order has nothing to do
		 * with a human language, it is intended for sendmail
		 * queue directories
		 */
		if (c1 != c2)
			return c1 - c2;
	}
	return CMP(i1,i2);
}

/* sort_group() return values (grouping order 1 < 2 < ... < 6 < 7) */
#define GROUP_DOTDIR	1
#define GROUP_DOTDOTDIR	2
#define GROUP_DIR		3
#define GROUP_BDEV		4
#define GROUP_CDEV		5
#define GROUP_OTHER		6
#define GROUP_PLAIN		7
static int
sort_group(FILE_ENTRY *pfe)
{
	int type;

	type = pfe->file_type;
	if (IS_FT_PLAIN(type))
		return GROUP_PLAIN;
	if (IS_FT_DIR(type)) {
		if (pfe->dotdir == 1)
			return GROUP_DOTDIR;
		if (pfe->dotdir == 2)
			return GROUP_DOTDOTDIR;
		return GROUP_DIR;
	}
	if (group_files == 2) {
		if (type == FT_DEV_CHAR)
			return GROUP_CDEV;
		if (type == FT_DEV_BLOCK)
			return GROUP_BDEV;
	}
	return GROUP_OTHER;
}

static int
qcmp(const void *e1, const void *e2)
{
	int cmp, group1, group2;
	const char *str1, *str2;
	time_t time1, time2;
	off_t size1, size2;

	group1 = 0;	/* to prevent compiler warning */

	/* I. file type grouping */
	if (group_files > 0) {
		group1 = sort_group(*(FILE_ENTRY **)e1);
		group2 = sort_group(*(FILE_ENTRY **)e2);
		if (group1 != group2)
			return group1 - group2;
	}

	/* special sorting for devices */
	if (group_files == 2
	  && (group1 == GROUP_BDEV || group1 == GROUP_CDEV) ) {
		/*
		 * numbers in 'sd_str.device' are properly formatted, i.e.
		 * first major, then minor, fixed position, left padded
		 */
		str1 = (*(FILE_ENTRY **)e1)->sd_str.device;
		str2 = (*(FILE_ENTRY **)e2)->sd_str.device;
		if ( (cmp = strcmp(str1,str2)) )
			return cmp;
	}

	/* II. sort order */
	switch (panel_sort.order) {
	case SORT_EMAN:
		str1 = SDSTR((*(FILE_ENTRY **)e1)->file);
		str2 = SDSTR((*(FILE_ENTRY **)e2)->file);
		return revstrcmp(str1,str2);
	case SORT_SUFFIX:
		str1 = (*(FILE_ENTRY **)e1)->extension;
		str2 = (*(FILE_ENTRY **)e2)->extension;
		cmp = STRCOLL(str1,str2);
		break;
	case SORT_TIME:
		time1 = (*(FILE_ENTRY **)e1)->mtime;
		time2 = (*(FILE_ENTRY **)e2)->mtime;
		cmp = CMP(time2,time1);
		break;
	case SORT_TIME_REV:
		time1 = (*(FILE_ENTRY **)e1)->mtime;
		time2 = (*(FILE_ENTRY **)e2)->mtime;
		cmp = CMP(time1,time2);
		break;
	case SORT_SIZE:
		size1 = (*(FILE_ENTRY **)e1)->size;
		size2 = (*(FILE_ENTRY **)e2)->size;
		cmp = CMP(size1,size2);
		break;
	case SORT_SIZE_REV:
		size1 = (*(FILE_ENTRY **)e1)->size;
		size2 = (*(FILE_ENTRY **)e2)->size;
		cmp = CMP(size2,size1);
		break;
	default:	/* SORT_NAME */
		cmp = 0;
	}
	if (cmp)
		return cmp;

	/* III. sort by file name */
	str1 = SDSTR((*(FILE_ENTRY **)e1)->file);
	str2 = SDSTR((*(FILE_ENTRY **)e2)->file);
	return STRCOLL(str1,str2);
}

void
sort_files(void)
{
	if (ppanel_file->cnt == 0)
		return;

	group_files = config_num(CFG_GROUP_FILES);
	qsort(ppanel_file->files,ppanel_file->cnt,
	  sizeof(FILE_ENTRY *),qcmp);
}
